'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019, 2024 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
from collections.abc import Iterable
import io
import pandas as pd
import numpy as np
from matplotlib.axes import Axes
from sklearn.tree import export_graphviz
try:
    from IPython.display import Image
except ImportError:
    Image = None
try:
    import pydotplus
except ImportError:
    pydotplus = None


def liftChart(data: pd.DataFrame, *, ranking: str | None = None, actual: str | None = None,
              title: str = 'Decile Lift Chart', labelBars: bool = True,
              ax: Axes | None = None, figsize: Iterable[float] = None):
    """ Create a decile lift chart using ranking and predicted values 

    Input: 
        data: DataFrame with ranking and actual values
        ranking: column name for ranking (predicted values or probability)
        actual: column name for actual values
        title (optional): set to None to suppress title
        labelBars (optional): set to False to avoid mean response labels on bar chart
        ax (optional): axis for matplotlib graph
        figsize (optional): size of matplotlib graph
    """
    if ranking is None or actual is None:
        raise ValueError('Column names for ranking and actual must be provided')

    data = data.sort_values(by=[ranking], ascending=False)
    ranked_actual = data[actual]

    # group the sorted actual values into 10 roughly equal groups and calculate the mean
    groups = [int(10 * i / len(ranked_actual))
              for i in range(len(ranked_actual))]
    meanPercentile = ranked_actual.groupby(groups).mean()
    # divide by the mean prediction to get the mean response
    meanResponse = meanPercentile / ranked_actual.mean()
    meanResponse.index = (meanResponse.index + 1) * 10

    ax = meanResponse.plot.bar(color='C0', ax=ax, figsize=figsize)
    ax.set_ylim(0, 1.12 * meanResponse.max() if labelBars else None)
    ax.set_xlabel('Percentile')
    ax.set_ylabel('Decile mean / global mean')
    if title:
        ax.set_title(title)

    if labelBars:
        for p in ax.patches:
            ax.annotate('{:.1f}'.format(p.get_height()),
                        (p.get_x(), p.get_height() + 0.1))
    return ax


def gainsChart(data: pd.DataFrame, *, ranking: str | None = None, actual: str | None = None,
               color: str = 'C0', title='Cumulative Gains Chart', label: str | None = None,
               ax: Axes | None = None, figsize: Iterable[float] | None = None):
    """ Create a gains chart using ranking and predicted values 

    Input: 
        data: DataFrame with ranking and actual values
        ranking: column name for ranking (predicted values or probability)
        actual: column name for actual values
        color (optional): color of graph
        title (optional): set to None to suppress title
        ax (optional): axis for matplotlib graph
        figsize (optional): size of matplotlib graph
    """
    if ranking is None or actual is None:
        raise ValueError('Column names for ranking and actual must be provided')
    data = data.sort_values(by=[ranking], ascending=False)
    gains = data[actual]

    nTotal = len(gains)  # number of records
    nActual = gains.sum()  # number of desired records

    # get cumulative sum of gains and convert to percentage
    # Note the additional 0 at the front
    cumGains = pd.concat([pd.Series([0]), gains.cumsum()])
    gains_df = pd.DataFrame(
        {'records': list(range(len(gains) + 1)), 'cumGains': cumGains})

    ax = gains_df.plot(x='records', y='cumGains', color=color, label=label, legend=False,
                       ax=ax, figsize=figsize)

    # Add line for random gain
    ax.plot([0, nTotal], [0, nActual], linestyle='--', color='k')
    ax.set_title(title)
    ax.set_xlabel('# records')
    ax.set_ylabel('# cumulative gains')
    return ax


def plotDecisionTree(decisionTree, feature_names=None, class_names=None, impurity=False, label='root',
                     max_depth=None, rotate=False, pdfFile=None):
    """ Create a plot of the scikit-learn decision tree and show in the Jupyter notebooke 
    Input:
        decisionTree: scikit-learn decision tree
        feature_names (optional): variable names
        class_names (optional): class names, only relevant for classification trees
        impurity (optional): show node impurity
        label (optional): only show labels at the root
        max_depth (optional): limit 
        rotate (optional): rotate the layout of the graph
        pdfFile (optional): provide pathname to create a PDF file of the graph
    """
    if pydotplus is None:
        return 'You need to install pydotplus to visualize decision trees'
    if Image is None:
        return 'You need to install ipython to visualize decision trees'
    if class_names is not None:
        class_names = [str(s) for s in class_names]  # convert to strings
    dot_data = io.StringIO()
    export_graphviz(decisionTree, feature_names=feature_names, class_names=class_names, impurity=impurity,
                    label=label, out_file=dot_data, filled=True, rounded=True, special_characters=True,
                    max_depth=max_depth, rotate=rotate)
    graph = pydotplus.graph_from_dot_data(dot_data.getvalue())
    if pdfFile is not None:
        graph.write_pdf(str(pdfFile))
    return Image(graph.create_png())

# Taken from scikit-learn documentation


def textDecisionTree(decisionTree, indent='  ', as_ratio=True):
    """ Create a text representation of the scikit-learn decision tree 
    Input:
        decisionTree: scikit-learn decision tree
        as_ratio: show the composition of the leaf nodes as ratio (default) instead of counts 
        indent: indentation (default two spaces)
    """
    n_nodes = decisionTree.tree_.node_count
    children_left = decisionTree.tree_.children_left
    children_right = decisionTree.tree_.children_right
    feature = decisionTree.tree_.feature
    threshold = decisionTree.tree_.threshold
    node_value = decisionTree.tree_.value

    node_depth = np.zeros(shape=n_nodes, dtype=np.int64)
    is_leaves = np.zeros(shape=n_nodes, dtype=bool)
    stack = [(0, -1)]  # seed is the root node id and its parent depth
    while len(stack) > 0:
        node_id, parent_depth = stack.pop()
        node_depth[node_id] = parent_depth + 1
        # If we have a test node
        if (children_left[node_id] != children_right[node_id]):
            stack.append((children_left[node_id], parent_depth + 1))
            stack.append((children_right[node_id], parent_depth + 1))
        else:
            is_leaves[node_id] = True

    rep = []
    for i in range(n_nodes):
        common = f'{node_depth[i] * indent}node={i}'
        if is_leaves[i]:
            value = node_value[i]
            if as_ratio:
                value = [[round(vi / sum(v), 3) for vi in v] for v in value]
            rep.append(f'{common} leaf node: {value}')
        else:
            rule = f'{children_left[i]} if {feature[i]} <= {threshold[i]} else to node {children_right[i]}'
            rep.append(f'{common} test node: go to node {rule}')
    return '\n'.join(rep)



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck
'''
import math
import numpy as np
from sklearn.metrics import r2_score, mean_squared_error
from sklearn.metrics import confusion_matrix, accuracy_score


def adjusted_r2_score(y_true, y_pred, model):
    """ calculate adjusted R2
    Input:
        y_true: actual values
        y_pred: predicted values
        model: predictive model
    """
    n = len(y_pred)
    p = len(model.coef_)
    if p >= n - 1:
        return 0
    r2 = r2_score(y_true, y_pred)
    return 1 - (1 - r2) * (n - 1) / (n - p - 1)


def AIC_score(*, y_true, y_pred, model=None, df=None):
    """ calculate Akaike Information Criterion (AIC) 
    Input:
        y_true: actual values
        y_pred: predicted values
        model (optional): predictive model
        df (optional): degrees of freedom of model

    One of model or df is requried
    """
    if df is None and model is None:
        raise ValueError('You need to provide either model or df')
    n = len(y_pred)
    p = len(model.coef_) + 1 if df is None else df
    if isinstance(y_true[0], str):
        sse = sum(yt != yp for yt, yp in zip(y_true, y_pred))
    else:
        resid = np.array(y_true) - np.array(y_pred)
        sse = np.sum(resid ** 2)
    constant = n + n * np.log(2 * np.pi)
    return n * math.log(sse / n) + constant + 2 * (p + 1)


def BIC_score(*, y_true, y_pred, model=None, df=None):
    """ calculate Schwartz's Bayesian Information Criterion (AIC) 
    Input:
        y_true: actual values
        y_pred: predicted values
        model: predictive model
        df (optional): degrees of freedom of model
    """
    aic = AIC_score(y_true=y_true, y_pred=y_pred, model=model, df=df)
    p = len(model.coef_) + 1 if df is None else df
    n = len(y_pred)
    return aic - 2 * (p + 1) + math.log(n) * (p + 1)


def regressionSummary(*, y_true, y_pred):
    """ print regression performance metrics 

    Input:
        y_true: actual values
        y_pred: predicted values
    """
    metrics = regressionMetrics(y_true=y_true, y_pred=y_pred)
    label = {
        'ME': 'Mean Error (ME)',
        'RMSE': 'Root Mean Squared Error (RMSE)',
        'MAE': 'Mean Absolute Error (MAE)',
        'MPE': 'Mean Percentage Error (MPE)',
        'MAPE': 'Mean Absolute Percentage Error (MAPE)',
    }
    fmt1 = '{{:>{}}} : {{:.4f}}'.format(max(len(m[0]) for m in metrics))
    print('\nRegression statistics\n')
    for metric, value in metrics.items():
        print(fmt1.format(label[metric], value))


def regressionMetrics(*, y_true, y_pred):
    """ calculate and return regression performance metrics 

    Input:
        y_true: actual values
        y_pred: predicted values
    Output:
        dictionary of regression metrics
    """
    y_true = _toArray(y_true)
    y_pred = _toArray(y_pred)
    y_res = y_true - y_pred
    metrics = {
        'ME': sum(y_res) / len(y_res),
        'RMSE': math.sqrt(mean_squared_error(y_true, y_pred)),
        'MAE': sum(abs(y_res)) / len(y_res),
    }
    if all(yt != 0 for yt in y_true):
        metrics['MPE'] = 100 * sum(y_res / y_true) / len(y_res)
        metrics['MAPE'] = 100 * sum(abs(y_res / y_true) / len(y_res))
    return metrics


def _toArray(y):
    y = np.asarray(y)
    if len(y.shape) == 2 and y.shape[1] == 1:
        y = y.ravel()
    return y


def classificationSummary(*, y_true, y_pred, class_names=None):
    """ Print a summary of classification performance

    Input:
        y_true: actual values
        y_pred: predicted values
        class_names (optional): list of class names
    """
    if class_names is None:
        class_names = sorted({*y_true, *y_pred})
    labels = class_names
    confusionMatrix = confusion_matrix(y_true, y_pred, labels=labels)
    accuracy = accuracy_score(y_true, y_pred)

    print('Confusion Matrix (Accuracy {:.4f})\n'.format(accuracy))

    # Pretty-print confusion matrix
    cm = confusionMatrix

    # Convert the confusion matrix and labels to strings
    cm = [[str(i) for i in row] for row in cm]
    labels = [str(i) for i in labels]

    # Determine the width for the first label column and the individual cells
    prediction = 'Prediction'
    actual = 'Actual'
    labelWidth = max(len(s) for s in labels)
    cmWidth = max(max(len(s) for row in cm for s in row), labelWidth) + 1
    labelWidth = max(labelWidth, len(actual))

    # Construct the format statements
    fmt1 = '{{:>{}}}'.format(labelWidth)
    fmt2 = '{{:>{}}}'.format(cmWidth) * len(labels)

    # And print the confusion matrix
    print(fmt1.format(' ') + ' ' + prediction)
    print(fmt1.format(actual), end='')
    print(fmt2.format(*labels))

    for cls, row in zip(labels, cm):
        print(fmt1.format(cls), end='')
        print(fmt2.format(*row))



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019, 2024 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
import urllib
import os
import pandas as pd


def printTermDocumentMatrix(count_vect, counts):
    """ Print term-document matrix created by the CountVectorizer 
    Input:
        count_vect: scikit-learn Count vectorizer
        counts: term-document matrix returned by transform method of counter vectorizer
    """
    shape = counts.shape
    columns = ['S{}'.format(i) for i in range(1, shape[0] + 1)]
    print(pd.DataFrame(data=counts.toarray().transpose(),
                       index=count_vect.get_feature_names_out(), columns=columns))


def downloadGloveModel():
    """ Download the GloVe model """
    if not os.path.exists('glove.6B.zip'):
        urllib.request.urlretrieve('http://nlp.stanford.edu/data/glove.6B.zip', 'glove.6B.zip')



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019, 2024 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
import os
import matplotlib as mpl
from .version import __version__

if os.environ.get('DISPLAY', '') == '' and os.name != 'nt':
    print('no display found. Using non-interactive Agg backend')
    mpl.use('Agg')

from .featureSelection import exhaustive_search, forward_selection, backward_elimination, stepwise_selection
from .graphs import plotDecisionTree, liftChart, gainsChart, textDecisionTree
from .metric import regressionSummary, classificationSummary, regressionMetrics
from .metric import AIC_score, BIC_score, adjusted_r2_score
from .textMining import downloadGloveModel, printTermDocumentMatrix
from .data import load_data, get_data_file



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019, 2024 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
import unittest
from io import StringIO
from contextlib import redirect_stdout

import pandas as pd
from sklearn.feature_extraction.text import CountVectorizer

from mlba import printTermDocumentMatrix


class TestTextMining(unittest.TestCase):
    def test_printTermDocumentMatrix(self):
        text = ['this is the first sentence.',
                'this is a second sentence.',
                'the third sentence is here.']
        count_vect = CountVectorizer()
        counts = count_vect.fit_transform(text)

        pd.set_option('display.width', 100)
        pd.set_option('display.max_columns', 20)

        out = StringIO()
        with redirect_stdout(out):
            printTermDocumentMatrix(count_vect, counts)
        s = out.getvalue()
        self.assertIn('S1  S2  S3', s)
        self.assertIn('first      1   0   0', s)
        self.assertIn('the        1   0   1', s)



'''
Created on Jun 12, 2020

@author: gedeck
'''
from pathlib import Path

import pandas as pd


DATA_DIR = Path(__file__).parent / 'csvFiles'


def load_data(name, **kwargs):
    """ Returns the data either as a Pandas data frame or series """
    data_file = get_data_file(name)
    if not data_file.exists():
        raise ValueError('Data file {name} not found')
    data = pd.read_csv(data_file, **kwargs)
    if data.shape[1] == 1:
        return data[data.columns[0]]
    return data


def get_data_file(name):
    if name.endswith('.zip'):
        return DATA_DIR / name
    if name.endswith('.gz'):
        name = name[:-3]
    if name.endswith('.csv'):
        name = name[:-4]
    return DATA_DIR / f'{name}.csv.gz'



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and
Applications in Python"

(c) 2019, 2024 Galit Shmueli, Peter C. Bruce, Peter Gedeck
'''
import unittest
from collections import namedtuple
from contextlib import redirect_stdout
from io import StringIO

import pandas as pd
from sklearn.metrics import r2_score

from mlba import adjusted_r2_score, AIC_score, BIC_score
from mlba import regressionSummary, classificationSummary


MockModel = namedtuple('Model', 'coef_')


class TestMetric(unittest.TestCase):
    def test_adjusted_r2_score(self):
        y_true = [1, 2, 3, 4, 5]
        y_pred = [1, 3, 2, 5, 4]

        r2 = r2_score(y_true, y_pred)
        assert r2 == 0.6

        for df in range(4):
            n = len(y_true)
            f = (n - 1) / (n - df - 1)
            expected = 1 - (1 - r2) * f
            self.assertAlmostEqual(
                adjusted_r2_score(y_true, y_pred, MockModel(coef_=[1] * df)),
                expected, places=3, msg=f'failed for df={df}')

        # if degree of freedom gets too large returns 0
        coef = [1] * 4
        self.assertAlmostEqual(
            adjusted_r2_score(y_true, y_pred, MockModel(coef_=coef)),
            0, places=3, msg=f'failed for df={df}')

        coef = [1] * 5
        self.assertAlmostEqual(
            adjusted_r2_score(y_true, y_pred, MockModel(coef_=coef)),
            0, places=3, msg=f'failed for df={df}')

    def test_AIC_score(self):
        y_true = [1, 2, 3, 4, 5]
        y_pred = [1, 3, 2, 5, 4]

        self.assertAlmostEqual(
            AIC_score(y_true=y_true, y_pred=y_pred, model=MockModel(coef_=[1] * 2)),
            21.0736, places=3)

        self.assertAlmostEqual(
            AIC_score(y_true=y_true, y_pred=y_pred, df=3),
            AIC_score(y_true=y_true, y_pred=y_pred, model=MockModel(coef_=[1] * 2)), places=3)

        self.assertGreater(
            AIC_score(y_true=y_true, y_pred=y_pred, df=3),
            AIC_score(y_true=y_true, y_pred=y_pred, df=2))

    def test_AIC_score_classification(self):
        y_true = ['b' if y == 1 else 'a' for y in [1, 0, 0, 1, 1, 1]]
        y_pred = ['b' if y == 1 else 'a' for y in [1, 0, 1, 1, 0, 0]]
        self.assertAlmostEqual(
            AIC_score(y_true=y_true, y_pred=y_pred, df=3),
            20.868379, places=3)

        self.assertAlmostEqual(
            AIC_score(y_true=pd.Series(y_true), y_pred=pd.Series(y_pred), df=3),
            20.868379, places=3)

    def test_BIC_score(self):
        y_true = [1, 2, 3, 4, 5]
        y_pred = [1, 3, 2, 5, 4]

        self.assertAlmostEqual(
            BIC_score(y_true=y_true, y_pred=y_pred, model=MockModel(coef_=[1] * 2)),
            19.51141, places=3)

        self.assertAlmostEqual(
            BIC_score(y_true=y_true, y_pred=y_pred, df=3),
            BIC_score(y_true=y_true, y_pred=y_pred, model=MockModel(coef_=[1] * 2)), places=3)

        self.assertGreater(
            BIC_score(y_true=y_true, y_pred=y_pred, df=3),
            BIC_score(y_true=y_true, y_pred=y_pred, df=2))

    def test_regressionSummary(self):
        y_true = [1, 2, 3, 4, 5]
        y_pred = [1, 3, 2, 5, 4]

        out = StringIO()
        with redirect_stdout(out):
            regressionSummary(y_true=y_true, y_pred=y_pred)
        s = out.getvalue()
        self.assertIn('Regression statistics', s)
        self.assertIn('(ME) : 0.0000', s)
        self.assertIn('(RMSE) : 0.8944', s)
        self.assertIn('(MAE) : 0.8000', s)
        self.assertIn('(MPE) : -4.3333', s)
        self.assertIn('(MAPE) : 25.6667', s)

        y_true = [0, 1, 2, 3, 4]
        y_pred = [0, 2, 1, 4, 3]

        out = StringIO()
        with redirect_stdout(out):
            regressionSummary(y_true=y_true, y_pred=y_pred)
        s = out.getvalue()
        self.assertIn('Regression statistics', s)
        self.assertIn('(ME) : 0.0000', s)
        self.assertIn('(RMSE) : 0.8944', s)
        self.assertIn('(MAE) : 0.8000', s)
        self.assertNotIn('(MPE)', s)
        self.assertNotIn('(MAPE)', s)

        y_true = [[1], [2], [3], [4], [5]]
        y_pred = [[1], [3], [2], [5], [4]]

        out = StringIO()
        with redirect_stdout(out):
            regressionSummary(y_true=y_true, y_pred=y_pred)
        s = out.getvalue()
        self.assertIn('Regression statistics', s)
        self.assertIn('(ME) : 0.0000', s)
        self.assertIn('(RMSE) : 0.8944', s)
        self.assertIn('(MAE) : 0.8000', s)
        self.assertIn('(MPE) : -4.3333', s)
        self.assertIn('(MAPE) : 25.6667', s)

    def test_classificationSummary(self):
        y_true = ['b' if y == 1 else 'a' for y in [1, 0, 0, 1, 1, 1]]
        y_pred = ['b' if y == 1 else 'a' for y in [1, 0, 1, 1, 0, 0]]

        out = StringIO()
        with redirect_stdout(out):
            classificationSummary(y_true=y_true, y_pred=y_pred, class_names=['a', 'b'])
        s = out.getvalue()

        self.assertIn('Confusion Matrix', s)
        self.assertIn('       Prediction', s)
        self.assertIn('a 1 1', s)
        self.assertIn('b 2 2', s)



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019, 2024 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
import itertools


def exhaustive_search(variables, train_model, score_model):
    """ Variable selection using backward elimination 

    Input:
        variables: complete list of variables to consider in model building
        train_model: function that returns a fitted model for a given set of variables
        score_model: function that returns the score of a model; better models have lower scores

    Returns:
        List of best subset models for increasing number of variables
    """
    # create models of increasing size and determine the best models in each case
    result = []
    for nvariables in range(1, len(variables) + 1):
        best_subset = None
        best_score = None
        best_model = None
        for subset in itertools.combinations(variables, nvariables):
            subset = list(subset)
            subset_model = train_model(subset)
            subset_score = score_model(subset_model, subset)
            if best_subset is None or best_score > subset_score:
                best_subset = subset
                best_score = subset_score
                best_model = subset_model
        result.append({
            'n': nvariables,
            'variables': best_subset,
            'score': best_score,
            'model': best_model,
        })
    return result


def backward_elimination(variables, train_model, score_model, verbose=False):
    """ Variable selection using backward elimination 

    Input:
        variables: complete list of variables to consider in model building
        train_model: function that returns a fitted model for a given set of variables
        score_model: function that returns the score of a model; better models have lower scores

    Returns:
        (best_model, best_variables) 
    """
    # we start with a model that contains all variables
    best_variables = list(variables)
    best_model = train_model(best_variables)
    best_score = score_model(best_model, best_variables)
    if verbose:
        print('Variables: ' + ', '.join(variables))
        print('Start: score={:.2f}'.format(best_score))

    while len(best_variables) > 1:
        step = [(best_score, None, best_model)]
        for removeVar in best_variables:
            step_var = list(best_variables)
            step_var.remove(removeVar)
            step_model = train_model(step_var)
            step_score = score_model(step_model, step_var)
            step.append((step_score, removeVar, step_model))

        # sort by ascending score
        step.sort(key=lambda x: x[0])

        # the first entry is the model with the lowest score
        best_score, removed_step, best_model = step[0]
        if verbose:
            print('Step: score={:.2f}, remove {}'.format(
                best_score, removed_step))
        if removed_step is None:
            # step here, as removing more variables is detrimental to performance
            break
        best_variables.remove(removed_step)
    return best_model, best_variables


def forward_selection(variables, train_model, score_model, verbose=True):
    """ Variable selection using forward selection 

    Input:
        variables: complete list of variables to consider in model building
        train_model: function that returns a fitted model for a given set of variables
        score_model: function that returns the score of a model; better models have lower scores

    Returns:
        (best_model, best_variables) 
    """
    # we start with a model that contains no variables
    best_variables = []
    best_model = train_model(best_variables)
    best_score = score_model(best_model, best_variables)
    if verbose:
        print('Variables: ' + ', '.join(variables))
        print('Start: score={:.2f}, constant'.format(best_score))
    while True:
        step = [(best_score, None, best_model)]
        for addVar in variables:
            if addVar in best_variables:
                continue
            step_var = list(best_variables)
            step_var.append(addVar)
            step_model = train_model(step_var)
            step_score = score_model(step_model, step_var)
            step.append((step_score, addVar, step_model))
        step.sort(key=lambda x: x[0])

        # the first entry in step is now the model that improved most
        best_score, added_step, best_model = step[0]
        if verbose:
            print('Step: score={:.2f}, add {}'.format(best_score, added_step))
        if added_step is None:
            # stop here, as adding more variables is detrimental to performance
            break
        best_variables.append(added_step)
    return best_model, best_variables


def stepwise_selection(variables, train_model, score_model, direction='both', verbose=True):
    """ Variable selection using forward and/or backward selection 

    Input:
        variables: complete list of variables to consider in model building
        train_model: function that returns a fitted model for a given set of variables
        score_model: function that returns the score of a model; better models have lower scores
        direction: use it to limit stepwise selection to either 'forward' or 'backward'

    Returns:
        (best_model, best_variables) 
    """
    FORWARD = 'forward'
    BACKWARD = 'backward'
    directions = [FORWARD, BACKWARD]
    if direction.lower() == FORWARD:
        directions = [FORWARD]
    if direction.lower() == BACKWARD:
        directions = [BACKWARD]

    # we start with a model that contains no variables
    best_variables = [] if 'forward' in directions else list(variables)
    best_model = train_model(best_variables)
    best_score = score_model(best_model, best_variables)
    if verbose:
        print('Variables: ' + ', '.join(variables))
        print('Start: score={:.2f}, constant'.format(best_score))

    while True:
        step = [(best_score, None, best_model, 'unchanged')]
        if FORWARD in directions:
            for variable in variables:
                if variable in best_variables:
                    continue
                step_var = list(best_variables)
                step_var.append(variable)
                step_model = train_model(step_var)
                step_score = score_model(step_model, step_var)
                step.append((step_score, variable, step_model, 'add'))

        if 'backward' in directions:
            for variable in best_variables:
                step_var = list(best_variables)
                step_var.remove(variable)
                step_model = train_model(step_var)
                step_score = score_model(step_model, step_var)
                step.append((step_score, variable, step_model, 'remove'))

        # sort by ascending score
        step.sort(key=lambda x: x[0])

        # the first entry is the model with the lowest score
        best_score, chosen_variable, best_model, direction = step[0]
        if verbose:
            print('Step: score={:.2f}, {} {}'.format(
                best_score, direction, chosen_variable))
        if chosen_variable is None:
            # step here, as adding or removing more variables is detrimental to performance
            break
        if direction == 'add':
            best_variables.append(chosen_variable)
        else:
            best_variables.remove(chosen_variable)
    return best_model, best_variables



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019, 2024 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
from pathlib import Path
import unittest

import pytest

from mlba.data import DATA_DIR
import mlba
import pandas as pd


class TestData(unittest.TestCase):
    def test_load_data(self):
        with pytest.raises(ValueError):
            mlba.load_data('unknown data file')

        for name in ('Amtrak.csv', ):
            data = mlba.load_data(name)
            assert isinstance(data, pd.DataFrame)

    def test_load_data_all(self):
        for name in Path(DATA_DIR).glob('*.csv.gz'):
            data = mlba.load_data(name.name)
            assert isinstance(data, (pd.Series, pd.DataFrame))
            assert len(data.shape) <= 2
            if len(data.shape) == 1:
                assert isinstance(data, pd.Series)
                print(name)
            else:
                assert isinstance(data, pd.DataFrame)
                assert data.shape[1] > 1

    def test_kwargs_load_data(self):
        df = mlba.load_data('gdp.csv')
        org_length = len(df)
        df = mlba.load_data('gdp.csv', skiprows=4)
        assert org_length == len(df) + 4

    def test_get_data_file(self):
        assert mlba.get_data_file('AutoAndElectronics.zip').exists()
        assert mlba.get_data_file('gdp.csv').exists()
        assert mlba.get_data_file('gdp.csv.gz').exists()



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019, 2024 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
__version__ = '1.9.24'



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck
'''
import unittest

import pandas as pd

from mlba import liftChart, gainsChart, textDecisionTree
from sklearn.datasets import load_iris
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection import train_test_split


class TestGraphs(unittest.TestCase):
    def test_liftChart(self):
        data = pd.Series([7] * 10 + [2.5] * 10 + [0.5]
                         * 10 + [0.25] * 20 + [0.1] * 50)
        df = pd.DataFrame({'ranking': data, 'actual': range(len(data))})
        ax = liftChart(df, actual='actual', ranking='ranking')
        self.assertIsNotNone(ax)

    def test_gainsChart(self):
        data = pd.Series([7] * 10 + [2.5] * 10 + [0.5]
                         * 10 + [0.25] * 20 + [0.1] * 50)
        df = pd.DataFrame({'ranking': data, 'actual': range(len(data))})
        ax = gainsChart(df, actual='actual', ranking='ranking')
        self.assertIsNotNone(ax)

    def test_textDecisionTree(self):
        iris = load_iris()
        X = iris.data
        y = iris.target
        X_train, _, y_train, _ = train_test_split(X, y, random_state=0)

        estimator = DecisionTreeClassifier(max_leaf_nodes=3, random_state=0)
        estimator.fit(X_train, y_train)

        representation = textDecisionTree(estimator)
        # print(representation)
        self.assertIn('node=0 test node', representation)
        self.assertIn('node=1 leaf node', representation)
        self.assertIn('node=2 test node', representation)
        self.assertIn('node=3 leaf node', representation)
        self.assertIn('node=4 leaf node', representation)
